@@ -228,7 +228,8 @@ public final class Binder {
         new byte[] { CallSiteDescriptor.ARG_OBJ, CallSiteDescriptor.ARG_STR }, null);
     private static final CallSiteDescriptor bindParamThrower = new CallSiteDescriptor(
         new byte[] { CallSiteDescriptor.ARG_OBJ, CallSiteDescriptor.ARG_OBJ,
-            CallSiteDescriptor.ARG_STR, CallSiteDescriptor.ARG_OBJ
+            CallSiteDescriptor.ARG_STR, CallSiteDescriptor.ARG_OBJ,
+            CallSiteDescriptor.ARG_INT
         }, null);
     private static final CallSiteDescriptor bindConcreteThrower = new CallSiteDescriptor(
         new byte[] { CallSiteDescriptor.ARG_STR, CallSiteDescriptor.ARG_STR,
@@ -238,7 +239,7 @@ public final class Binder {
     private static final CallSiteDescriptor paramReadWriteThrower = new CallSiteDescriptor(
         new byte[] { CallSiteDescriptor.ARG_OBJ, CallSiteDescriptor.ARG_STR }, null);
     private static int bindOneParam(ThreadContext tc, RakOps.GlobalExt gcx, CallFrame cf, SixModelObject param,
-            Object origArg, byte origFlag, boolean noNomTypeCheck, Object[] error) {
+            Object origArg, byte origFlag, boolean noNomTypeCheck, boolean isSlurpyNamed, Object[] error) {
         /* Get parameter flags and variable name. */
         param.get_attribute_native(tc, gcx.Parameter, "$!flags", HINT_flags);
         int paramFlags = (int)tc.native_i;
@@ -377,6 +378,8 @@ public final class Binder {
         SixModelObject decontValue = null;
         boolean didHLLTransform = false;
         SixModelObject nomType = null;
+        SixModelObject ContextRef = null;
+        SixModelObject HOW = null;
         if (flag == CallSiteDescriptor.ARG_OBJ && !(is_rw && desiredNative != 0)) {
             /* We need to work on the decontainerized value. */
             decontValue = Ops.decont(arg_o, tc);
@@ -395,10 +398,10 @@ public final class Binder {
                 nomType = param.get_attribute_boxed(tc, gcx.Parameter,
                     "$!nominal_type", HINT_nominal_type);
                 if ((paramFlags & SIG_ELEM_NOMINAL_GENERIC) != 0) {
-                    SixModelObject HOW = nomType.st.HOW;
+                    HOW = nomType.st.HOW;
                     SixModelObject ig = Ops.findmethod(HOW,
                         "instantiate_generic", tc);
-                    SixModelObject ContextRef = tc.gc.ContextRef;
+                    ContextRef = tc.gc.ContextRef;
                     SixModelObject cc = ContextRef.st.REPR.allocate(tc, ContextRef.st);
                     ((ContextRefInstance)cc).context = cf;
                     Ops.invokeDirect(tc, ig, genIns,
@@ -409,13 +412,13 @@ public final class Binder {
                 /* If the expected type is Positional, see if we need to do the
                  * positional bind failover. */
                 if (nomType == gcx.Positional) {
-                    if (Ops.istype_nodecont(arg_o, gcx.PositionalBindFailover, tc) != 0) {
+                    if (Ops.istype_nd(arg_o, gcx.PositionalBindFailover, tc) != 0) {
                         SixModelObject ig = Ops.findmethod(arg_o, "cache", tc);
                         Ops.invokeDirect(tc, ig, Ops.invocantCallSite, new Object[] { arg_o });
                         arg_o = Ops.result_o(tc.curFrame);
                         decontValue = Ops.decont(arg_o, tc);
                     }
-                    else if (Ops.istype_nodecont(decontValue, gcx.PositionalBindFailover, tc) != 0) {
+                    else if (Ops.istype_nd(decontValue, gcx.PositionalBindFailover, tc) != 0) {
                         SixModelObject ig = Ops.findmethod(decontValue, "cache", tc);
                         Ops.invokeDirect(tc, ig, Ops.invocantCallSite, new Object[] { decontValue });
                         decontValue = Ops.result_o(tc.curFrame);
@@ -423,17 +426,18 @@ public final class Binder {
                 }
 
                 /* If not, do the check. If the wanted nominal type is Mu, then
-                 * anything goes. */
-                if (nomType != gcx.Mu && Ops.istype_nodecont(decontValue, nomType, tc) == 0) {
+                 * anything goes.
+                 * When binding a slurpy named hash while compiling the setting don't check for Associative.
+                 */
+                if (nomType != gcx.Mu && !(isSlurpyNamed && nomType == gcx.Associative) && Ops.istype_nd(decontValue, nomType, tc) == 0) {
                     /* Type check failed; produce error if needed. */
                     if (error != null) {
-
                         SixModelObject thrower = RakOps.getThrower(tc, "X::TypeCheck::Binding::Parameter");
                         if (thrower != null) {
                             error[0] = thrower;
                             error[1] = bindParamThrower;
                             error[2] = new Object[] { decontValue, nomType.st.WHAT,
-                                varName, param };
+                                varName, param, (long)0 };
                         }
                         else {
                             error[0] = String.format(
@@ -511,10 +515,29 @@ public final class Binder {
                 return BIND_RESULT_FAIL;
             }
 
+            /* Is the coercion target generic and in need of instantiation?
+             * (This can happen in (::T, T) where we didn't learn about the
+             * type until during the signature bind.) */
+            param.get_attribute_native(tc, gcx.Parameter, "$!coerce_method", HINT_coerce_method);
+            String methName = tc.native_s;
+            HOW = coerceType.st.HOW;
+            SixModelObject archetypesMeth = Ops.findmethod(HOW, "archetypes", tc);
+            Ops.invokeDirect(tc, archetypesMeth, Ops.invocantCallSite, new Object[] { HOW });
+            SixModelObject Archetypes = Ops.result_o(tc.curFrame);
+            SixModelObject genericMeth = Ops.findmethod(Archetypes, "generic", tc);
+            Ops.invokeDirect(tc, genericMeth, Ops.invocantCallSite, new Object[] { Archetypes });
+            if (Ops.istrue(Ops.result_o(tc.curFrame), tc) == 1) {
+                ContextRef = tc.gc.ContextRef;
+                SixModelObject ctcc = ContextRef.st.REPR.allocate(tc, ContextRef.st);
+                ((ContextRefInstance)ctcc).context = cf;
+                SixModelObject ctig = Ops.findmethod(HOW, "instantiate_generic", tc);
+                Ops.invokeDirect(tc, ctig, genIns, new Object[] { HOW, coerceType, ctcc });
+                coerceType = Ops.result_o(tc.curFrame);
+                methName = Ops.typeName(coerceType, tc);
+            }
+
             /* Only coerce if we don't already have the correct type. */
             if (Ops.istype(decontValue, coerceType, tc) == 0) {
-                param.get_attribute_native(tc, gcx.Parameter, "$!coerce_method", HINT_coerce_method);
-                String methName = tc.native_s;
                 SixModelObject coerceMeth = Ops.findmethodNonFatal(decontValue, methName, tc);
                 if (coerceMeth != null) {
                     Ops.invokeDirect(tc, coerceMeth,
@@ -648,7 +671,7 @@ public final class Binder {
                 /* Check we meet the constraint. */
                 SixModelObject consType = postConstraints.at_pos_boxed(tc, i);
                 SixModelObject acceptsMeth = Ops.findmethod(consType, "ACCEPTS", tc);
-                if (Ops.istype(consType, gcx.Code, tc) != 0)
+                if (Ops.isconcrete(consType, tc) == 1 && Ops.istype(consType, gcx.Code, tc) != 0)
                     RakOps.p6capturelex(consType, tc);
                 switch (flag) {
                     case CallSiteDescriptor.ARG_INT:
@@ -668,11 +691,22 @@ public final class Binder {
                             ACCEPTS_o, new Object[] { consType, arg_o });
                         break;
                 }
-                long result = Ops.istrue(
-                    Ops.result_o(tc.curFrame), tc);
-                if (result == 0) {
-                    if (error != null)
-                        error[0] = "Constraint type check failed for parameter '" + varName + "'";
+                if (Ops.istrue(Ops.result_o(tc.curFrame), tc) == 0) {
+                    /* Constraint type check failed; produce error if needed. */
+                    if (error != null) {
+                        SixModelObject thrower = RakOps.getThrower(tc, "X::TypeCheck::Binding::Parameter");
+                        if (thrower != null) {
+                            error[0] = thrower;
+                            error[1] = bindParamThrower;
+                            error[2] = new Object[] { (SixModelObject)origArg,
+                                consType.st.WHAT, varName, param, (long)1 };
+                        }
+                        else {
+                            error[0] = String.format(
+                                "Constraint type check failed for parameter '%s'",
+                                varName);
+                        }
+                    }
                     return BIND_RESULT_FAIL;
                 }
             }
@@ -874,7 +908,7 @@ public final class Binder {
                     capSnap.bind_attribute_boxed(tc, capType, "%!hash", HINT_CAPTURE_hash, namedArgs);
 
                     bindFail = bindOneParam(tc, gcx, cf, param, capSnap, CallSiteDescriptor.ARG_OBJ,
-                        noNomTypeCheck, error);
+                        noNomTypeCheck, false, error);
                 }
                 if (bindFail != 0) {
                     return bindFail;
@@ -900,7 +934,7 @@ public final class Binder {
                 bindee.bind_attribute_boxed(tc, gcx.Map, "$!storage",
                     HINT_ENUMMAP_storage, slurpy);
                 bindFail = bindOneParam(tc, gcx, cf, param, bindee, CallSiteDescriptor.ARG_OBJ,
-                    noNomTypeCheck, error);
+                    noNomTypeCheck, true, error);
                 if (bindFail != 0)
                     return bindFail;
 
@@ -943,7 +977,7 @@ public final class Binder {
                     SixModelObject bindee = Ops.result_o(tc.curFrame);
 
                     bindFail = bindOneParam(tc, gcx, cf, param, bindee, CallSiteDescriptor.ARG_OBJ,
-                        noNomTypeCheck, error);
+                        noNomTypeCheck, false, error);
                     if (bindFail != 0)
                         return bindFail;
                 }
@@ -954,7 +988,7 @@ public final class Binder {
                     if (curPosArg < numPosArgs) {
                         /* Easy - just bind that. */
                         bindFail = bindOneParam(tc, gcx, cf, param, args[curPosArg],
-                            csd.argFlags[curPosArg], noNomTypeCheck, error);
+                            csd.argFlags[curPosArg], noNomTypeCheck, false, error);
                         if (bindFail != 0)
                             return bindFail;
                         curPosArg++;
@@ -966,7 +1000,7 @@ public final class Binder {
                         if ((flags & SIG_ELEM_IS_OPTIONAL) != 0) {
                             bindFail = bindOneParam(tc, gcx, cf, param,
                                 handleOptional(tc, gcx, flags, param, cf),
-                                CallSiteDescriptor.ARG_OBJ, false, error);
+                                CallSiteDescriptor.ARG_OBJ, false, false, error);
                             if (bindFail != 0)
                                 return bindFail;
                         }
@@ -1000,7 +1034,7 @@ public final class Binder {
                     if ((flags & SIG_ELEM_IS_OPTIONAL) != 0) {
                         bindFail = bindOneParam(tc, gcx, cf, param,
                             handleOptional(tc, gcx, flags, param, cf),
-                            CallSiteDescriptor.ARG_OBJ, false, error);
+                            CallSiteDescriptor.ARG_OBJ, false, false, error);
                     }
                     else if (!suppressArityFail) {
                         if (error != null) {
@@ -1014,7 +1048,7 @@ public final class Binder {
                 }
                 else {
                     bindFail = bindOneParam(tc, gcx, cf, param, args[lookup >> 3],
-                        (byte)(lookup & 7), noNomTypeCheck, error);
+                        (byte)(lookup & 7), noNomTypeCheck, false, error);
                 }
 
                 /* If we got a binding failure, return it. */
